
#include "predef.h" 
#include <stdio.h>
#include <ctype.h>
#include <startnet.h>
#include <tcp.h>
#include <iointernal.h>
#include <iosys.h>
#include <serial.h>
#include <ucos.h>
#include "dualstdio.h"

#define RX_BUFFER_SIZE (2048)
	
static int serial_fd;
static int listen_fd;
static int telnet_fd;
static int our_extra_fd;
static char receive_buffer[RX_BUFFER_SIZE];
static volatile int rx_put;
static volatile int rx_get;
static OS_CRIT dual_crit;

IoExpandStruct our_expand_funcs;
static bool bShare;





int dual_read( int fd, char *buf, int nbytes )
{
int nbuf=0;
//Lock the critical section returning from function will clear this up as
//the object destructor will be called.
OSCriticalSectionObj crit(dual_crit);

while ((nbuf<nbytes) && (rx_get!=rx_put))
 {
 buf[nbuf++]=receive_buffer[rx_get++];
 if(rx_get==RX_BUFFER_SIZE) rx_get=0;
 }

if(rx_get==rx_put) ClrDataAvail(our_extra_fd); 
return nbuf;
}






int dual_write( int fd, const char *buf, int nbytes )
{
if(telnet_fd) 
	{
	 int rv=write(telnet_fd,buf,nbytes);
	 if(rv<=0) 
		 {close(telnet_fd);
		  telnet_fd=0;
		 }
	}

if((!telnet_fd)|| (bShare)) 
		writeall(serial_fd,buf,nbytes);

return nbytes;
}


int dual_close(int fd)
{
//Do nothing we really don't let you close stdio
return 0;
}


int ReadDualDataFrom(int fd)
{
 OSCriticalSectionObj crit(dual_crit);
 int nread=0;
 //I could be more efficent and read things in larger buffered sets by doing the math around the buffer
 //This is just stdio so I'm going to be really lazy and read one char at a time.
 
 while(dataavail(fd))
 {char buf[1];
	 int rv=read(fd,buf,1);
	 if(rv) 
		{ nread++;
		  receive_buffer[rx_put++]=buf[0];
		  if(rx_put==RX_BUFFER_SIZE) rx_put=0;
		  if(rx_get==rx_put) 
			{//We have wrapped around so bump rx_get loosing the oldes chars in the buffer
			  rx_get++;
			  if(rx_get==RX_BUFFER_SIZE) rx_get=0;
			}
		}
 }
if(nread) SetDataAvail(our_extra_fd); 

return nread;
}

void DualStdioTask(void * p)
{
while(1)
  {
	fd_set fd_read;
	fd_set fd_error;
	FD_ZERO(&fd_read);
	FD_ZERO(&fd_error);

	FD_SET(listen_fd,&fd_read);
	if(telnet_fd) 
		{
		FD_SET(telnet_fd,&fd_read);
		FD_SET(telnet_fd,&fd_error); 
		}
	FD_SET(serial_fd,&fd_read);

	select(32,&fd_read,(fd_set *) NULL,&fd_error,0);
	if(telnet_fd)
		{
        if(FD_ISSET(telnet_fd,&fd_error))
		{//We got an error
			close(telnet_fd);
			telnet_fd=0;
		}
		else
        if(FD_ISSET(telnet_fd,&fd_read))
		{
			if(ReadDualDataFrom(telnet_fd)==0)
				{//Read zero bytes so we have been closed
				 close(telnet_fd);
				 telnet_fd=0;
				}
		}

   	  }
	
	
	if(FD_ISSET(listen_fd,&fd_read))
	    {
		 int new_fd=accept(listen_fd,NULL,NULL,5);
		 if(new_fd)
		 {
			 //For now we just bump off the old telnet fd 
			 if(telnet_fd) 
			 {
				close(telnet_fd);
				telnet_fd=0;
			 }
		  telnet_fd=new_fd;
		 }
		}

	if(FD_ISSET(serial_fd,&fd_read))
	   {
		if((!telnet_fd)|| (bShare)) //No telnet or sharing read chars...
			ReadDualDataFrom(serial_fd);
		else
			{//We have a valid telnet connection and we are set to not share so 
			//we need to toss chars...
			//If we have mroe than 80 chars to read the select will com by again.
			char buffer[80];
			read(serial_fd,buffer,80);
			}
 	   }


  }
}


void CoreDualStdio(int serial_port,int baud_rate, WORD Telnet_port, BYTE prio, BOOL share)
{
if(serial_port<0) serial_port=gConfigRec.ser_boot;
if(baud_rate==0) baud_rate=gConfigRec.baud_rate;
if(Telnet_port==0) Telnet_port=23;


SerialClose(serial_port);
serial_fd=SimpleOpenSerial(serial_port,baud_rate);
listen_fd=listen(INADDR_ANY,Telnet_port);

OSCritInit(&dual_crit); 

our_expand_funcs.read= dual_read;
our_expand_funcs.write=dual_write;
our_expand_funcs.close=dual_close;
our_extra_fd=GetExtraFD(NULL,&our_expand_funcs);

bShare=share;

ClrDataAvail(our_extra_fd); 
SetWriteAvail(our_extra_fd); 
ClrHaveError(our_extra_fd); 


ReplaceStdio(0,our_extra_fd);
ReplaceStdio(1,our_extra_fd);
ReplaceStdio(2,our_extra_fd);

#ifdef OSSimpleTaskCreatewName
OSSimpleTaskCreatewName(DualStdioTask,prio,"Dual STDIO"); 
#else
OSSimpleTaskCreate(DualStdioTask,prio);
#endif
OSTimeDly(2);

}



//Set up so all stdio will go in/out of the telnet port if active and the serial port if not.
//setting baudrate to zero will cause it to use the config baud rate
//setting the telnet port to 0 will default to telnet (23)
//Only one or the other is active at any one time.
void StartSwappingDualStdio(int serial_port,int baud_rate, WORD Telnet_port, BYTE prio)
{
CoreDualStdio(serial_port,baud_rate,Telnet_port,prio,false);


}


//Set up so all stdio will go in/out of the telnet port if active and the serial port if not.
//setting baudrate to zero will cause it to use the config baud rate
//setting the telnet port to 0 will default to telnet (23)
//both the serial port and the telnet prot can send chars..
void StartSharingDualStdio(int serial_port, int baud_rate,WORD Telnet_port, BYTE prio)
{
CoreDualStdio(serial_port,baud_rate,Telnet_port,prio,true);
}

